/**
 * @file usb.c
 *
 * @brief This file contains the USB stack for the I2C-MP-USB firmware project.
 *
 * @author Thomas Fischl
 * @copyright (c) 2017-2019 Thomas Fischl
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include <xc.h>
#include "usb_descr.h"
#include "usb.h"

volatile EndpointType ep[EP_MAX] @ 0x2000;
volatile unsigned char ep0out_buffer[EP_BUFFERSIZE] @ 0x2020;
volatile unsigned char ep0in_buffer[EP_BUFFERSIZE] @ 0x2028;

unsigned short usb_sendleft = 0;
unsigned char usb_usertransfer = 0;
unsigned short usb_usertransfer_length = 0;

const unsigned char * usb_sendbuffer;
unsigned configured = 0;
unsigned char usb_config = 0;
unsigned char usb_setaddress = 0;
unsigned char usb_ep0status[2];

/**
 * @brief Initialize USB stack
 */
void usb_init() {
    ep[0].out.stat = 0x80;
    ep[0].out.cnt = EP_BUFFERSIZE;
    ep[0].out.adrl = 0x20;
    ep[0].out.adrh = 0x20;

    ep[0].in.stat = 0;
    ep[0].in.cnt = EP_BUFFERSIZE;
    ep[0].in.adrl = 0x28;
    ep[0].in.adrh = 0x20;

    UEP0 = 0x16; // enable IN, enable OUT, enable CONTROL, enable handshake

    UCFG = 0x14; // enable pullup and full-speed
    UCON = 0x08; // enable usb module

    usb_ep0status[0] = 0;
    usb_ep0status[1] = 0;
}

/**
 * @brief Shutdown USB stack
 */
void usb_shutdown() {
    UCON = 0x00;
    UCFG = 0x00;
}

/**
 * @brief Determine if serial number is available
 * @return 0 if no number available
 */
unsigned char usb_serialNumberAvailable() {
    return ((usb_string_serial[0] == USB_STRING_SERIALNUMBER_SIZE) && (usb_string_serial[1] == 0x03));
}

/**
 * @brief Process pending send activity
 */
void usb_sendProcess() {
    if (usb_sendleft == 0) return;

    unsigned short length = usb_sendleft;
    if (length > EP_BUFFERSIZE)
        length = EP_BUFFERSIZE;

    unsigned char i;
    for (i = 0; i < length; i++) {

        if ((usb_sendbuffer == usb_dev_desc + USB_DEV_DESC_SERIALNUMBER_OFFSET) && usb_serialNumberAvailable()) ep0in_buffer[i] = USB_STRING_SERIALNUMBER_INDEX;
        else ep0in_buffer[i] = *usb_sendbuffer;

        usb_sendbuffer++;
        usb_sendleft--;
    }

    ep[0].in.cnt = length;
    if (ep[0].in.stat & 0x40)
        ep[0].in.stat = 0x88;
    else
        ep[0].in.stat = 0xC8;
}

/**
 * @brief Load given descriptor buffer into send buffer
 *
 * @param descbuffer Pointer to descriptor
 * @param size Size of descriptor
 * @param length Count of bytes to send
 * @return 0 on error, 1 on success
 */
unsigned char usb_loadDescriptor(const unsigned char * descbuffer, unsigned short size, unsigned short length) {

    if (length > size)
        length = size;

    usb_sendleft = length;
    usb_sendbuffer = descbuffer;

    usb_sendProcess();
    return length;
}

/**
 * @brief Handle descriptor requests
 *
 * @param type Type of descriptor
 * @param index Index of descriptor
 * @param length Requested length
 * @return 0 on error, 1 on success
 */
unsigned char usb_handleDescriptorRequest(unsigned char type, unsigned char index, unsigned short length) {

    switch (type) {
        case DESCR_DEVICE:
            return usb_loadDescriptor(usb_dev_desc, sizeof (usb_dev_desc), length);
        case DESCR_CONFIG:
            return usb_loadDescriptor(usb_config_desc, sizeof (usb_config_desc), length);
        case DESCR_STRING:
            switch (index) {
                case 0: return usb_loadDescriptor(usb_string_language, sizeof (usb_string_language), length);
                case 1: return usb_loadDescriptor(usb_string_manuf, sizeof (usb_string_manuf), length);
                case 2: return usb_loadDescriptor(usb_string_product, sizeof (usb_string_product), length);
                case 3: return usb_loadDescriptor(usb_string_serial, USB_STRING_SERIALNUMBER_SIZE, length);
            }
    }

    return 0;
}

/**
 * @brief Do IN transfer
 */
void ut_in() {

    unsigned short length = usb_usertransfer_length;
    if (length > EP_BUFFERSIZE)
        length = EP_BUFFERSIZE;

    length = usb_vendor_in((unsigned char*) ep0in_buffer, length);

    usb_usertransfer_length -= length;

    ep[0].in.cnt = length;
    if (ep[0].in.stat & 0x40)
        ep[0].in.stat = 0x88;
    else
        ep[0].in.stat = 0xC8;

    if (length < EP_BUFFERSIZE) {
        usb_usertransfer = 0;
    }
}

/**
 * @brief Do USB stack processing
 */
void usb_process() {


    if (UIRbits.TRNIF) {
        // complete interrupt

        if (USTAT == USTAT_EP0_OUT) {

            // out/setup

            if (((ep[0].out.stat >> 2) & 0x0F) == USB_PID_SETUP) {
                // setup token

                usb_usertransfer = 0;

                // stop pending/old/prepared IN transactions
                ep[0].in.stat = 0;
                usb_sendleft = 0;

                if ((ep0out_buffer[0] & USBRQ_TYPE_MASK) == USBRQ_TYPE_STANDARD) {

                    switch (ep0out_buffer[1]) {
                        case REQUEST_GET_DESCRIPTOR:
                            if (!usb_handleDescriptorRequest(ep0out_buffer[3], ep0out_buffer[2], (ep0out_buffer[7] << 8) | ep0out_buffer[6])) {
                                ep[0].in.cnt = 0;
                                ep[0].in.stat = 0xCC; // Stall
                            }
                            break;
                        case REQUEST_SET_ADDRESS:

                            usb_setaddress = ep0out_buffer[2];

                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;

                            break;

                        case REQUEST_SET_CONFIGURATION:

                            usb_config = ep0out_buffer[2];
                            configured = 1;
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_GET_CONFIGURATION:

                            ep0in_buffer[0] = usb_config;
                            ep[0].in.cnt = 1;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_GET_INTERFACE:

                            ep0in_buffer[0] = 0;
                            ep[0].in.cnt = 1;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_GET_STATUS:
                            if ((ep0out_buffer[0] & USBRQ_RECIPIENT_MASK) == USBRQ_RECIPIENT_ENDPOINT) {
                                ep0in_buffer[0] = usb_ep0status[0];
                                ep0in_buffer[1] = usb_ep0status[1];
                            } else {
                                ep0in_buffer[0] = 0;
                                ep0in_buffer[1] = 0;
                            }
                            ep[0].in.cnt = 2;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_SET_FEATURE:
                            if ((ep0out_buffer[0] & USBRQ_RECIPIENT_MASK) == USBRQ_RECIPIENT_ENDPOINT) {
                                if (ep0out_buffer[2] == 0x00) // HALT
                                    usb_ep0status[0] = 1;
                            }
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_CLEAR_FEATURE:
                            if ((ep0out_buffer[0] & USBRQ_RECIPIENT_MASK) == USBRQ_RECIPIENT_ENDPOINT) {
                                usb_ep0status[0] = 0;
                            }
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_SYNCH_FRAME:
                            ep0in_buffer[0] = 0;
                            ep0in_buffer[1] = 0;
                            ep[0].in.cnt = 2;
                            ep[0].in.stat = 0xC8;
                            break;

                        case REQUEST_SET_INTERFACE:
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;
                            break;

                        default:
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xCC; // stall
                            break;

                    }
                } else {

                    // not standard type -> class or vendor

                    if ((ep0out_buffer[0] & 0x80) == 0x80) {
                        // vendor or class control in

                        unsigned char len = usb_vendor_setup((unsigned char *) ep0out_buffer, (unsigned char *) ep0in_buffer);

                        if (len == 0xff) {
                            usb_usertransfer_length = (ep0out_buffer[7] << 8) | ep0out_buffer[6];
                            usb_usertransfer = 1;
                            ut_in();
                        } else {
                            ep[0].in.cnt = len;
                            ep[0].in.stat = 0xC8;
                        }


                    } else {
                        // vendor or class control out

                        unsigned char len = usb_vendor_setup((unsigned char *) ep0out_buffer, (unsigned char *) ep0in_buffer);
                        if (len == 0xff) {
                            usb_usertransfer_length = (ep0out_buffer[7] << 8) | ep0out_buffer[6];
                            usb_usertransfer = 2;
                        } else {
                            // TODO: what to do here? is zero package correct?
                            ep[0].in.cnt = 0;
                            ep[0].in.stat = 0xC8;
                        }
                    }
                }

            } else {
                // USB_PID_OUT

                // stop pending/old/prepared IN transactions
                ep[0].in.stat = 0;
                usb_sendleft = 0;

                if (usb_usertransfer == 2) {

                    unsigned char ack = usb_vendor_out((unsigned char*) ep0out_buffer, ep[0].out.cnt);

                    if (ack) {
                        ep[0].in.cnt = 0;
                        ep[0].in.stat = 0xC8;
                    } else {
                        ep[0].in.cnt = 0;
                        ep[0].in.stat = 0xCC; // Stall
                    }
                }

            }

            ep[0].out.cnt = EP_BUFFERSIZE;
            ep[0].out.stat = 0x80;


        } else if (USTAT == USTAT_EP0_IN) {

            // check if set address command is pending
            if (usb_setaddress > 0) {
                UADDR = usb_setaddress;
                usb_setaddress = 0;
            }

            if (usb_usertransfer == 1) {
                ut_in();
            } else {
                usb_sendProcess();
            }
        }

        UCONbits.PKTDIS = 0;
        UIRbits.TRNIF = 0;
    }
}
